import uuid from 'uuid/v4';
import _ from 'lodash';
import { datasourceApi } from '../../../services/datasource';

const prefix = uuid();

const getKeyName = (meta) => {
  const keyProperty = _(meta.properties || [])
    .filter(property => !property.skip)
    .find(property => property.key);
  return keyProperty ? keyProperty.name : undefined;
};

const getSource = (property) => {
  if (property.source) {
    return (property.source.items || []).map((item) => {
      try {
        return JSON.parse(item);
      } catch (err) {
        return item;
      }
    });
  } else {
    return [];
  }
};

const parseMetas = (metas) => {
  const ret = {};
  if (!metas) {
    return ret;
  }
  _.forEach(metas, (value, key) => {
    let finalValue;
    try {
      finalValue = JSON.parse(value);
    } catch (err) {
      finalValue = value;
    }
    _.set(ret, key, finalValue);
  });
  return _.pickBy(ret, _.negate(_.isUndefined));
};

const makeProps = (meta) => {
  if (!meta || !meta.metas) {
    return {};
  }
  const props = parseMetas(meta.metas);
  if (!props.rowKey) {
    props.rowKey = getKeyName(meta);
  }
  return props;
};

const makeColumns = (meta) => {
  return (meta.properties || [])
    .filter(property => !property.skip)
    .map((property) => {
      const props = parseMetas(property.metas);
      if (!props.title && props.label) {
        props.title = props.label;
      }
      if (props.order === undefined) {
        props.order = 0;
      }
      if ((props.fixed === true || props.fixed === 'left' || props.fixed === 'right') && props.width === undefined) {
        props.width = 150;
      }
      return _.pickBy({
        ...props,
        dataIndex: property.name,
        key: property.name,
        sorter: property.sort,
        filterType: property.filterType,
        filterEnums: getSource(property),
      }, _.negate(_.isUndefined));
    })
    .filter(c => c.visible !== false)
    .sort((c1, c2) => {
      const c1Left = c1.fixed === true || c1.fixed === 'left';
      const c1Right = c1.fixed === 'right';
      const c2Left = c2.fixed === true || c2.fixed === 'left';
      const c2Right = c2.fixed === 'right';
      if (c1Left && !c2Left) {
        return -1;
      }
      if (!c1Left && c2Left) {
        return 1;
      }
      if (c1Right && !c2Right) {
        return 1;
      }
      if (!c1Right && c2Right) {
        return -1;
      }
      return c1.order - c2.order
    });
};

const getArrayData = ({ dataType, arrayData, singularData }, meta) => {
  if (dataType === 'TABLE') {
    const data = (arrayData || []).map(() => ({}));
    (meta.properties || [])
      .filter(property => !property.skip)
      .forEach((property, i) => {
        data.forEach((row, j) => {
          row[property.name] = arrayData[j][i]; // eslint-disable-line no-param-reassign
        });
      });
    return data;
  } else if (dataType === 'PROPERTIES') {
    const data = [];
    (meta.properties || [])
      .filter(property => !property.skip)
      .forEach((property) => {
        data.push((singularData || {})[property.name]);
      });
    return [_.toPlainObject(data)];
  } else {
    throw new Error(`Unsupported data type: ${dataType}`);
  }
};

const modelCreator = () => {
  const name = 'model';
  const namespace = _.uniqueId(prefix);
  return {
    name,
    model: {
      namespace,
      state: {
        columns: [],
        props: {},
        num: 0,
        list: [],
      },
      reducers: {
        queryMetaSuccess(state, { payload: { props, columns } }) {
          return {
            ...state,
            props,
            columns,
          };
        },
        queryCountSuccess(state, { payload: num }) {
          return {
            ...state,
            num,
          };
        },
        queryTasksSuccess(state, { payload: list }) {
          return {
            ...state,
            list,
          };
        },
      },
      effects: {
        *fetchData({ payload: { coordinate, pst, psz, params, filters} }, { put, call }) {
          const options = { pst, psz, params, filters };
          const api = datasourceApi(coordinate);
          const meta = yield call(api.meta);
          const props = makeProps(meta);
          yield put({ type: 'queryMetaSuccess', payload: { props, columns: makeColumns(meta) } });
          const num = yield call(api.count, options);
          yield put({ type: 'queryCountSuccess', payload: num });
          const dsb = yield call(api.query, options);
          yield put({ type: 'queryTasksSuccess', payload: getArrayData(dsb, meta) });
        },
      },
      subscriptions: {},
    },
  };
};

export default modelCreator;
